import numpy as np
from polliwog import Polyline

DEFAULT_RADIUS = 1.0


def add_landmark_points(mesh, coords, radius=DEFAULT_RADIUS):
    offset = radius * np.eye(3)
    segments = np.repeat(coords, 6, axis=0).reshape(-1, 3, 2, 3)
    segments[:, :, 0] = segments[:, :, 0] + offset
    segments[:, :, 1] = segments[:, :, 1] - offset
    polylines = [
        Polyline(v=segment_vs, is_closed=False)
        for segment_vs in segments.reshape(-1, 2, 3)
    ]
    mesh.add_lines(polylines)



"""
On Mac OS:

.. code-block:: sh

    brew install spatialindex
    pip install rtree trimesh
"""
from __future__ import print_function


def require_trimesh_with_rtree():
    """
    Check that trimesh and rtree are installed and can be imported, and raise
    an error with a helpful error message if they are not.
    """
    try:
        import rtree
        import trimesh

        # For flake8.
        assert rtree
        assert trimesh
    except ImportError:  # pragma: no cover
        print(
            """
            libspatialindex, rtree, and trimesh are not installed. On Mac OS:

            $ brew install spatialindex
            $ pip install rtree trimesh
            """
        )
        raise ImportError("trimesh, rtree, and spatialindex are not installed")


def faces_nearest_to_points(mesh, query_points, ret_points=False):
    """
    Find the triangular faces on a mesh which are nearest to the given query
    points.

    Args:
        query_points (np.arraylike): The points to query, with shape `kx3`
        ret_points (bool): When `True`, return both the indices of the
            nearest faces and the closest points to the query points, which
            are not necessarily vertices. When `False`, return only the
            face indices.

    Returns:
        object: face indices as `kx1 np.ndarray`, or when `ret_points`
        is `True`, a tuple also including the coordinates of the closest points
        as `kx3 np.ndarray`.
    """
    import trimesh

    trimesh_mesh = trimesh.Trimesh(vertices=mesh.v, faces=mesh.f)
    closest_points, _, face_indices = trimesh_mesh.nearest.on_surface(query_points)
    return (face_indices, closest_points) if ret_points else face_indices



import numpy as np


def shuffle_vertices(mesh):
    """
    Shuffle the mesh's vertex ordering, preserving the integrity of the faces.
    The mesh is mutated.

    Args:
        mesh (lace.mesh.Mesh): A mesh.

    Returns:
        np.ndarray: `vx1` mapping of old vertex indices to new.
    """
    v_old_to_new = np.random.permutation(len(mesh.v))
    mesh.reorder_vertices(v_old_to_new)
    return v_old_to_new


def shuffle_faces(mesh):
    """
    Shuffle the mesh's face ordering. The mesh is mutated.

    Args:
        mesh (lace.mesh.Mesh): A mesh.

    Returns:
        np.ndarray: `fx1` mapping of old face indices to new.
    """
    f_old_to_new = np.random.permutation(len(mesh.f))
    mesh.f = mesh.f[f_old_to_new]
    return f_old_to_new



from cached_property import cached_property
import numpy as np
from .landmarker import Landmarker
from .landmark_compositor import LandmarkCompositor
from ._mesh import DEFAULT_RADIUS


class LandmarkCompositeRecipe(object):
    def __init__(self, recipe):
        """
        Example recipe:

        base_mesh: examples/average.obj
        landmarks:
          - knee_left
          - knee_right
        examples:
          - id: example01
            mesh: examples/example01.obj
            knee_left: [-10.0, 15.0, 4.0]
            knee_right: [10.0, 14.8, 4.1]
          - id: example02
            mesh: examples/example02.obj
            knee_left: [-11.0, 13.0, 3.5]
            knee_right: [12.0, 12.8, 3.4]
        """
        self.base_mesh_path = recipe["base_mesh"]
        self.landmark_names = recipe["landmarks"]
        self.examples = recipe["examples"]

    @classmethod
    def load(cls, recipe_path):
        import yaml

        with open(recipe_path, "r") as f:
            recipe_data = yaml.load(f)
        return cls(recipe_data)

    @property
    def base_mesh(self):
        from lace.mesh import Mesh

        return Mesh(filename=self.base_mesh_path)

    @cached_property
    def composite_landmarks(self):
        from lace.mesh import Mesh

        compositor = LandmarkCompositor(
            base_mesh=self.base_mesh, landmark_names=self.landmark_names
        )
        for example in self.examples:
            example_mesh = Mesh(filename=example["mesh"])
            landmarks = {k: example[k] for k in self.landmark_names}
            compositor.add_example(mesh=example_mesh, landmarks=landmarks)
        return compositor.result

    @cached_property
    def reprojected_landmarks(self):
        from lace.mesh import Mesh

        inverse_landmarker = Landmarker(
            source_mesh=self.base_mesh, landmarks=self.composite_landmarks
        )

        reprojected = {}
        for example in self.examples:
            mesh = Mesh(filename=example["mesh"])
            reprojected[example["id"]] = inverse_landmarker.transfer_landmarks_onto(
                mesh
            )
        return reprojected

    def write_reprojected_landmarks(self, output_dir, radius=DEFAULT_RADIUS):
        import os
        from lace.mesh import Mesh
        from ._mesh import add_landmark_points

        for example in self.examples:
            mesh = Mesh(filename=example["mesh"])
            example_id = example["id"]

            original_landmarks = np.array([example[k] for k in self.landmark_names])
            reprojected_landmarks = np.array(
                list(self.reprojected_landmarks[example_id].values())
            )
            add_landmark_points(
                mesh,
                np.vstack([original_landmarks, reprojected_landmarks]),
                radius=radius,
            )
            # Make the original landmarks blue, and the reprojected landmarks
            # the default red.
            e_color = [
                {
                    "color": (0, 0, 1),
                    "e_indices": np.arange(len(mesh.e) / 2, dtype=np.uint64),
                }
            ]

            out = os.path.join(output_dir, "{}.dae".format(example_id))
            mesh.write_dae(out, e_color)



from setuptools import setup, find_packages

# Set version_info[__version__], while avoiding importing numpy, in case numpy
# and vg are being installed concurrently.
# https://packaging.python.org/guides/single-sourcing-package-version/
version_info = {}
exec(open("entente/package_version.py").read(), version_info)

readme = open("README.md", "rb").read().decode("utf-8")
install_requires = open("requirements.txt", "rb").read().decode("utf-8")

setup(
    name="entente",
    version=version_info["__version__"],
    description="Work with polygonal meshes which have vertex-wise correspondence",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Metabolize",
    author_email="github@paulmelnikow.com",
    url="https://github.com/lace/entente",
    project_urls={
        "Issue Tracker": "https://github.com/lace/entente/issues",
        "Documentation": "https://entente.readthedocs.io/en/stable/",
    },
    packages=find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Artistic Software",
        "Topic :: Multimedia :: Graphics :: 3D Modeling",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)



def composite_meshes(mesh_paths):
    """
    Create a composite as a vertex-wise average of several meshes in
    correspondence. Faces, groups, and other attributes are loaded from the
    first mesh given.

    Args:
        mesh_paths (list): Paths of the meshes to average.

    Returns:
        lace.mesh.Mesh: The composite mesh.
    """
    from lace.mesh import Mesh

    if not len(mesh_paths):
        raise ValueError("Expected at least one mesh path")

    first_mesh_path, remaining_mesh_paths = mesh_paths[0], mesh_paths[1:]

    working_mesh = Mesh(filename=first_mesh_path)

    for this_mesh_path in remaining_mesh_paths:
        this_mesh = Mesh(filename=this_mesh_path)
        if not this_mesh.has_same_topology(working_mesh):
            raise ValueError(
                "Expected {} to have the same topology as {}".format(
                    this_mesh_path, first_mesh_path
                )
            )
        working_mesh.v += this_mesh.v

    working_mesh.v /= len(mesh_paths)

    return working_mesh
